from unittest import TestCase
from tensorflow.keras.models import Model
import tensorflow.keras.layers as layers

import orcanet.builder_util.layer_blocks as layer_blocks


class TestInceptionBlockV2(TestCase):
    def setUp(self):
        self.inp_2d = layers.Input(shape=(10, 10, 1))
        self.inp_3d = layers.Input(shape=(10, 10, 10, 1))

    def test_shape_2d(self):
        x = layer_blocks.InceptionBlockV2(
            conv_dim=2,
            filters_pool=3,
            filters_1x1=4,
            filters_3x3=(5, 6),
            filters_3x3dbl=(7, 8),
        )(self.inp_2d)
        self.assertSequenceEqual(
            Model(self.inp_2d, x).output_shape, (None, 10, 10, 21))

    def test_shape_2d_strides(self):
        x = layer_blocks.InceptionBlockV2(
            conv_dim=2,
            filters_pool=3,
            filters_1x1=4,
            filters_3x3=(5, 6),
            filters_3x3dbl=(7, 8),
            strides=2,
        )(self.inp_2d)
        self.assertSequenceEqual(
            Model(self.inp_2d, x).output_shape, (None, 5, 5, 15))

    def test_shape_3d(self):
        x = layer_blocks.InceptionBlockV2(
            conv_dim=3,
            filters_pool=3,
            filters_1x1=4,
            filters_3x3=(5, 6),
            filters_3x3dbl=(7, 8),
            strides=1,
        )(self.inp_3d)
        self.assertSequenceEqual(
            Model(self.inp_3d, x).output_shape, (None, 10, 10, 10, 21))

    def test_shape_3d_strides(self):
        x = layer_blocks.InceptionBlockV2(
            conv_dim=3,
            filters_pool=3,
            filters_1x1=4,
            filters_3x3=(5, 6),
            filters_3x3dbl=(7, 8),
            strides=2,
        )(self.inp_3d)
        self.assertSequenceEqual(
            Model(self.inp_3d, x).output_shape, (None, 5, 5, 5, 15))

    def test_2d_params(self):
        x = layer_blocks.InceptionBlockV2(
            conv_dim=2,
            filters_pool=3,
            filters_1x1=4,
            filters_3x3=(5, 6),
            filters_3x3dbl=(7, 8),
        )(self.inp_2d)
        self.assertEqual(
            Model(self.inp_2d, x).count_params(), 450)

    def test_2d_params_stride(self):
        x = layer_blocks.InceptionBlockV2(
            conv_dim=2,
            filters_pool=3,
            filters_1x1=4,
            filters_3x3=(5, 6),
            filters_3x3dbl=(7, 8),
            strides=2,
        )(self.inp_2d)
        self.assertEqual(
            Model(self.inp_2d, x).count_params(), 436)


class TestConvBlock(TestCase):
    def setUp(self):
        self.inp_2d = layers.Input(shape=(9, 10, 1))
        self.inp_3d = layers.Input(shape=(8, 9, 10, 1))

    def test_2d(self):
        x = layer_blocks.ConvBlock(
            conv_dim=2,
            filters=11,
            dropout=0.2,
            pool_size=2,
        )(self.inp_2d)
        model = Model(self.inp_2d, x)

        self.assertEqual(model.count_params(), 110)
        self.assertSequenceEqual(model.output_shape, (None, 4, 5, 11))
        target_layers = (
            layers.InputLayer,
            layers.Conv2D,
            layers.Activation,
            layers.MaxPool2D,
            layers.Dropout,
        )
        for layer, target_layer in zip(model.layers, target_layers):
            self.assertIsInstance(layer, target_layer)

    def test_2d_time_distributed(self):
        x = layer_blocks.ConvBlock(
            conv_dim=2,
            filters=11,
            dropout=0.2,
            pool_size=2,
            time_distributed=True,
            padding=[1, 0],
        )(self.inp_3d)
        model = Model(self.inp_3d, x)

        self.assertEqual(model.count_params(), 110)
        self.assertSequenceEqual(model.output_shape, (None, 8, 4, 4, 11))
        target_layers = [layers.InputLayer] + [layers.TimeDistributed] * 5

        for layer, target_layer in zip(model.layers, target_layers):
            self.assertIsInstance(layer, target_layer)



"""
Some basic sample modifiers to use with orcanet.
Use them by setting .cfg.sample_modifier of the orcanet.core.Organizer.

"""
from abc import abstractmethod
import numpy as np


class BaseModifier:
    """
    Parent class for modifiers that do the same operation on each input.
    Apply modify on x_value of each input, and output as dict.

    """
    def __call__(self, info_blob):
        x_values = info_blob["x_values"]
        xs = dict()
        for key, x_value in x_values.items():
            xs[key] = self.modify(x_value)
        return xs

    @abstractmethod
    def modify(self, x_value):
        """ x_value is a batch of input data as a numpy array. """
        raise NotImplementedError


class Permute(BaseModifier):
    """
    Permute the axes of the samples to given order.
    Batchsize axis is excluded, i.e. start indexing with 1!

    Example
    -------
    organizer.cfg.sample_modifier = Permute((2, 1, 3))
    --> Swap first two axes of each sample.

    """
    def __init__(self, axes):
        self.axes = list(axes)

    @classmethod
    def from_str(cls, string):
        """ E.g. '1,0,2' --> [1, 0, 2] """
        return cls([int(i) for i in string.split(",")])

    def modify(self, x_value):
        return np.transpose(x_value, [0] + self.axes)


class Reshape(BaseModifier):
    """
    Reshape samples to given shape.
    Batchsize axis is excluded!

    Example
    -------
    organizer.cfg.sample_modifier = Reshape((11, 13, 18))
    --> Reshape each sample to that shape.
    
    """
    def __init__(self, newshape):
        self.newshape = list(newshape)

    @classmethod
    def from_str(cls, string):
        """ E.g. '11,13,18' --> [11, 13, 18] """
        return cls([int(i) for i in string.split(",")])

    def modify(self, x_value):
        return np.reshape(x_value, [x_value.shape[0]] + self.newshape)


class JoinedModifier(BaseModifier):
    """
    For applying multiple sample modifiers after each other.

    Example
    -------
    organizer.cfg.sample_modifier = JoinedModifier([
        Reshape((11, 13, 18)), Permute((2, 1, 3))
    ])
    --> Reshape each sample, then permute axes.

    """
    def __init__(self, sample_modifiers):
        self.sample_modifiers = sample_modifiers

    def modify(self, x_value):
        result = x_value
        for smod in self.sample_modifiers:
            result = smod.modify(result)
        return result
